module phoenixCore (
    input extClk,
    input reset,

    inout [7:0] data_bus,
    inout [15:0] address_bus,

    input interrupt_trigger,
    output interrupt_done,

    output SOFTWARE_INTERRUPT,

    output mem_write,
    output mem_read,

    output IO_OUT, // DONE
    output IO_IN // DONE
);

// ----------------------- states -----------------------
    localparam STATE_FETCH_0 = 3'd0;
    localparam STATE_FETCH_1 = 3'd1;
    localparam STATE_FETCH_2 = 3'd2;
    localparam STATE_EXECUTE_MC = 3'd3;

    localparam STATE_INTERRUPT_PREPARE = 3'd4;
// ----------------------- states -----------------------



// ----------------------- controll lines -----------------------
    wire PUT_A_ADDR; // DONE
    wire PUT_IR0_DB; // DONE
    wire PUT_IR1_DB; // DONE
    
    wire SAVE_DB_IR0; // DONE
    wire SAVE_DB_IR1; // DONE
    
    wire PUT_IIMM_DB; // DONE

    wire SAVE_DB_ALUA; // DONE
    wire SAVE_DB_ALUB; // DONE

    wire ALU_ADD; // DONE
    wire ALU_CMP; // DONE

    wire PUT_ALUOUT_DB; // DONE

    wire COND_INVERT; // DONE

    wire SAVE_PCB_IF_Z; // DONE
    wire SAVE_PCB_IF_EQ; // DONE
    wire SAVE_PCB; // DONE

    wire SAVE_DB_MEM; // DONE
    wire PUT_MEM_DB; // DONE

    wire PC_FLUSH; // DONE
    wire PC_FLUSH_COND; // DONE

    wire ALU_SUB; // DONE
    wire ALU_NAND; // DONE
    wire ALU_NOR; // DONE

    wire PUT_B_ADDR; // DONE
    wire PUT_IADDR_ADDR; // DONE

    wire SAVE_ADDR_A; // DONE
    wire SAVE_ADDR_B; // DONE

    wire INT_RET_TO_ADDR; // DONE

    wire FINISH; // DONE

    wire SAVE_PCB_IF_OF; // DONE

    wire CLEAR_FG; // DONE

    wire ALU_ADO; // DONE
    wire ALU_SBO; // DONE

    wire SAVE_IH; // DONE

    wire PUT_FG_DB; // DONE
    wire PUT_DB_FG; // DONE

    wire SP_UP; // DONE
    wire SP_DOWN; // DONE

    wire SAVE_ADDR_SP; // DONE
    wire PUT_SP_ADDR; // DONE
// ----------------------- controll lines -----------------------


// ----------------------- internal registers -----------------------
    reg [15:0] pc;
    reg [15:0] pc_buffer;
    reg conditinal_buffer;

    reg [15:0] interrupt_handler;
    reg [15:0] interrupt_return_pc;
    reg interrupt_lock;

    reg [2:0] state;
    reg [2:0] mc_counter;

    reg [7:0] i0;
    reg [7:0] i1;
    reg [7:0] i2;

    wire [3:0] ir1 = i1[7:4];
    wire [3:0] ir0 = i1[3:0];

    wire [7:0] fg;

    wire halt = pc == 16'hffff || fg[3];
    wire clk = extClk && !halt;
// ----------------------- internal registers -----------------------


    phoenixMicrocode microcode (
        .clk(clk),
        .mask(state == STATE_EXECUTE_MC),
        .PUT_A_ADDR(PUT_A_ADDR),
        .PUT_IR0_DB(PUT_IR0_DB),
        .PUT_IR1_DB(PUT_IR1_DB),
        .SAVE_DB_IR0(SAVE_DB_IR0),
        .SAVE_DB_IR1(SAVE_DB_IR1),
        .PUT_IIMM_DB(PUT_IIMM_DB),
        .SAVE_DB_ALUA(SAVE_DB_ALUA),
        .SAVE_DB_ALUB(SAVE_DB_ALUB),
        .ALU_ADD(ALU_ADD),
        .ALU_CMP(ALU_CMP),
        .PUT_ALUOUT_DB(PUT_ALUOUT_DB),
        .COND_INVERT(COND_INVERT),
        .SAVE_PCB_IF_Z(SAVE_PCB_IF_Z),
        .SAVE_PCB_IF_EQ(SAVE_PCB_IF_EQ),
        .SAVE_PCB(SAVE_PCB),
        .SAVE_DB_MEM(SAVE_DB_MEM),
        .PUT_MEM_DB(PUT_MEM_DB),
        .PC_FLUSH(PC_FLUSH),
        .PC_FLUSH_COND(PC_FLUSH_COND),
        .ALU_SUB(ALU_SUB),
        .ALU_NAND(ALU_NAND),
        .ALU_NOR(ALU_NOR),
        .IO_IN(IO_IN),
        .IO_OUT(IO_OUT),
        .PUT_B_ADDR(PUT_B_ADDR),
        .PUT_IADDR_ADDR(PUT_IADDR_ADDR),
        .SAVE_ADDR_A(SAVE_ADDR_A),
        .SAVE_ADDR_B(SAVE_ADDR_B),
        .INT_RET_TO_ADDR(INT_RET_TO_ADDR),
        .SOFTWARE_INTERRUPT(SOFTWARE_INTERRUPT),
        .FINISH(FINISH),
        .SAVE_PCB_IF_OF(SAVE_PCB_IF_OF),
        .CLEAR_FG(CLEAR_FG),
        .ALU_ADO(ALU_ADO),
        .ALU_SBO(ALU_SBO),
        .SAVE_IH(SAVE_IH),
        .PUT_FG_DB(PUT_FG_DB),
        .PUT_DB_FG(PUT_DB_FG),
        .SP_UP(SP_UP),
        .SP_DOWN(SP_DOWN),
        .SAVE_ADDR_SP(SAVE_ADDR_SP),
        .PUT_SP_ADDR(PUT_SP_ADDR),
        .instruction(i0),
        .step(mc_counter)
    );

    phoenixAlu alu (
        .clk(clk),
        .reset(reset),
        .SAVE_DB_ALUA(SAVE_DB_ALUA),
        .SAVE_DB_ALUB(SAVE_DB_ALUB),
        .PUT_ALUOUT_DB(PUT_ALUOUT_DB),
        .ALU_ADD(ALU_ADD),
        .ALU_SUB(ALU_SUB),
        .ALU_NAND(ALU_NAND),
        .ALU_NOR(ALU_NOR),
        .ALU_ADO(ALU_ADO),
        .ALU_SBO(ALU_SBO),
        .ALU_CMP(ALU_CMP),
        .PUT_FG_DB(PUT_FG_DB),
        .PUT_DB_FG(PUT_DB_FG),
        .CLEAR_FG(CLEAR_FG),
        .data_bus(data_bus),
        .fg(fg)
    );

    phoenixRegisters registers (
        .clk(clk),
        .PUT_A_ADDR(PUT_A_ADDR),
        .PUT_B_ADDR(PUT_B_ADDR),
        .PUT_IR0_DB(PUT_IR0_DB),
        .PUT_IR1_DB(PUT_IR1_DB),
        .SAVE_ADDR_A(SAVE_ADDR_A),
        .SAVE_ADDR_B(SAVE_ADDR_B),
        .SAVE_DB_IR0(SAVE_DB_IR0),
        .SAVE_DB_IR1(SAVE_DB_IR1),
        .data_bus(data_bus),
        .address_bus(address_bus),
        .ir0(ir0),
        .ir1(ir1)
    );

    phoenixStackRegister stack_register (
        .clk(clk),
        .SP_UP(SP_UP),
        .SP_DOWN(SP_DOWN),
        .SAVE_ADDR_SP(SAVE_ADDR_SP),
        .PUT_SP_ADDR(PUT_SP_ADDR),
        .address_bus(address_bus)
    );

    assign mem_write = SAVE_DB_MEM;

    assign mem_read = (
        state == STATE_FETCH_0 ||
        state == STATE_FETCH_1 ||
        state == STATE_FETCH_2 ||
        PUT_MEM_DB
    );

    wire [15:0] address_bus_int = (
        (state == STATE_FETCH_0 || state == STATE_FETCH_1 || state == STATE_FETCH_2) ? pc : 16'd0 |
        PUT_IADDR_ADDR ? {i2, i1} : 16'd0 |
        INT_RET_TO_ADDR ? interrupt_return_pc : 16'd0
    );
    assign address_bus = (
        state == STATE_FETCH_0 ||
        state == STATE_FETCH_1 ||
        state == STATE_FETCH_2 ||
        PUT_IADDR_ADDR ||
        INT_RET_TO_ADDR
    ) ? address_bus_int : 16'bzzzzzzzzzzzzzzzz;


    assign data_bus = (PUT_IIMM_DB) ? i2 : 8'bzzzzzzzz;

    always @(negedge clk) begin
        if (SAVE_IH) begin
            interrupt_handler <= address_bus;
        end
    end
    assign interrupt_done = INT_RET_TO_ADDR;

    always @(negedge clk or posedge reset) begin
        if (reset) begin
            state <= STATE_FETCH_0;
            mc_counter <= 3'b000;
            pc <= 16'd0;
            interrupt_lock <= 1'b0;
        end else begin

            case (state)
                STATE_FETCH_0:
                    begin
                        if (interrupt_trigger && !interrupt_lock) begin
                            interrupt_lock <= 1'b1;
                            interrupt_return_pc <= pc;
                            state <= STATE_INTERRUPT_PREPARE;
                        end else begin
                            i0 = data_bus;
                            pc <= pc + 1;
                            state <= STATE_FETCH_1;
                        end
                    end
                STATE_FETCH_1:
                    begin
                        i1 = data_bus;
                        pc <= pc + 1;
                        state <= STATE_FETCH_2;
                    end
                STATE_FETCH_2:
                    begin
                        i2 = data_bus;
                        pc <= pc + 1;
                        state <= STATE_EXECUTE_MC;
                        mc_counter <= 3'd0;
                    end
                STATE_EXECUTE_MC:
                    begin
                        if (mc_counter == 3'b111) begin
                            state <= STATE_FETCH_0;
                        end
                        mc_counter = mc_counter + 1;
                    end
                STATE_INTERRUPT_PREPARE:
                    begin
                        pc <= interrupt_handler;
                        state <= STATE_FETCH_0;
                    end
            endcase

            if (FINISH) begin
                state <= STATE_FETCH_0;
            end

            if (INT_RET_TO_ADDR) begin
                interrupt_lock <= 1'b0;
            end

            if (SAVE_PCB) begin
                pc_buffer <= address_bus;
            end

            if (
                (SAVE_PCB_IF_Z && (fg[0] ^ COND_INVERT)) ||
                (SAVE_PCB_IF_EQ && (fg[1] ^ COND_INVERT)) ||
                (SAVE_PCB_IF_OF && (fg[2] ^ COND_INVERT))
            ) begin
                pc_buffer <= address_bus;
                conditinal_buffer <= 1'b1;
            end

            if (PC_FLUSH) begin
                pc <= pc_buffer;
            end

            if (PC_FLUSH_COND && conditinal_buffer) begin
                pc <= pc_buffer;
                conditinal_buffer <= 1'b0;
            end
        end
    end

    initial begin
        // TODO: improve
        $monitor("r0: %x, r1: %x, r2: %x, r3: %x, r4: %x, r5: %x, alua: %x, alub: %x, aluout: %x, fg: %b, halt: %b", registers.registers[0], registers.registers[1], registers.registers[2], registers.registers[3], registers.registers[4], registers.registers[5], alu.a, alu.b, alu.out, alu.fg, halt);
    end
endmodule