module phoenixUart (
    input clk,
    input reset,

    input io_in,
    input io_out,

    input [15:0] address_bus,
    inout [7:0] data_bus,

    output tx_pin,
    input rx_pin,

    output reg uart_int,

    input uart_clk
);

    wire [7:0] sendData;
    wire sendFinish;


    phoenixIo sendReg (
        .clk(clk),
        .reset(reset),
        .io_in(io_in),
        .io_out(io_out),
        .address_bus(address_bus),
        .data_bus(data_bus),
        .port_in(sendData),
        .port_out(sendData),
        .clear(sendFinish)
    );
    defparam sendReg.address = 16'hf000;

    uart_tx tx (
        .clk(uart_clk),
        .reset(reset),
        .din(sendData),
        .en(sendData != 8'd0),
        .tx(tx_pin),
        .finish(sendFinish)
    );



    wire [7:0] recvData;
    wire recvFinish;

    phoenixIo recvReg (
        .clk(clk),
        .reset(reset),
        .io_in(io_in),
        .io_out(io_out),
        .address_bus(address_bus),
        .data_bus(data_bus),
        .port_in(recvData),
        .clear(1'b0)
    );
    defparam recvReg.address = 16'hf001;

    uart_rx rx (
        .clk(uart_clk),
        .reset(reset),
        .rx(rx_pin),
        .dout(recvData),
        .finish(recvFinish)
    );

    always @(posedge clk or posedge recvFinish) begin
        if (recvFinish)
            uart_int <= 1'b1;
        else
            uart_int <= 1'b0;
    end
endmodule

module uart_rx (
    input clk,
    input reset,
    input rx,
    output reg [7:0] dout,
    output reg finish
);
    parameter CLK_FREQ = 27000000;
    parameter UART_FREQ = 115200;

    localparam STATE_IDLE   = 2'b00;
    localparam STATE_START  = 2'b01;
    localparam STATE_DATA   = 2'b10;
    localparam STATE_STOP   = 2'b11;

    reg [7:0] data = 8'h00;
    reg [2:0] bitpos = 3'h0;
    reg [1:0] state = STATE_IDLE;

    localparam RX_CLK_MAX = (CLK_FREQ / UART_FREQ) - 1;
    reg[$clog2(RX_CLK_MAX + 1) + 1:0] rx_clkcnt;
    wire rx_clk = (rx_clkcnt == 0);

    always @(posedge clk or posedge reset) begin
        if (reset) begin
            rx_clkcnt <= 0;
        end else begin
            if (rx_clkcnt >= RX_CLK_MAX) begin
                rx_clkcnt <= 0;
            end else begin
                rx_clkcnt <= rx_clkcnt + 1;
            end
        end
    end

    always @(posedge clk or posedge reset) begin
        if (reset) begin
            dout <= 8'h00;
            finish <= 1'b0;
        end else begin
            case (state)
                STATE_IDLE:
                    begin
                        finish <= 1'b0;
                        if (rx == 1'b0) begin
                            state <= STATE_START;
                            bitpos <= 3'h0;
                        end
                    end
                STATE_START:
                    begin
                        if (rx_clk) begin
                            if (rx == 1'b0) begin
                                state <= STATE_DATA;
                            end else begin
                                state <= STATE_IDLE;
                            end
                        end
                    end
                STATE_DATA:
                    begin
                        if (rx_clk) begin
                            if (bitpos == 3'h7)
                                state <= STATE_STOP;
                            else
                                bitpos <= bitpos + 3'h1;
                            data[bitpos] <= rx;
                        end
                    end
                STATE_STOP:
                    begin
                        if (rx_clk) begin
                            if (rx == 1'b1) begin
                                dout <= data;
                                state <= STATE_IDLE;
                                finish <= 1'b1;
                            end
                        end
                    end
            endcase
        end
    end
endmodule


module uart_tx (
    input clk,
    input reset,
    input [7:0] din,
    input en,
    output reg tx,
    output reg finish
);
    parameter clk_freq = 27000000;
    parameter uart_freq = 115200;

    localparam STATE_IDLE	= 2'b00;
    localparam STATE_START	= 2'b01;
    localparam STATE_DATA	= 2'b10;
    localparam STATE_STOP	= 2'b11;


    reg [7:0] data = 8'h00;
    reg [2:0] bitpos = 3'h0;
    reg [1:0] state = STATE_IDLE;


    localparam TX_CLK_MAX = (clk_freq / uart_freq) - 1;
    reg[$clog2(TX_CLK_MAX + 1) + 1:0] tx_clkcnt;
    wire tx_clk = (tx_clkcnt == 0);

    always @(posedge clk or posedge reset) begin
        if (reset) begin
            tx_clkcnt <= 0;
        end else begin
            if (tx_clkcnt >= TX_CLK_MAX) begin
                tx_clkcnt <= 0;
            end else begin
                tx_clkcnt <= tx_clkcnt + 1;
            end
        end
    end
        

    always @(posedge clk or posedge reset) begin
        if (reset) begin
            tx <= 1'b1;
            finish <= 1'b0;
        end else begin
            case (state)
                STATE_IDLE:
                    begin
                        finish <= 1'b0;
                        if (en) begin
                            state <= STATE_START;
                            data <= din;
                            bitpos <= 3'h0;
                        end
                    end
                STATE_START:
                    begin
                        if (tx_clk) begin
                            tx <= 1'b0;
                            state <= STATE_DATA;
                        end
                    end
                STATE_DATA:
                    begin
                        if (tx_clk) begin
                            if (bitpos == 3'h7)
                                state <= STATE_STOP;
                            else
                                bitpos <= bitpos + 3'h1;
                            tx <= data[bitpos];
                        end
                    end
                STATE_STOP:
                    begin
                        if (tx_clk) begin
                            tx <= 1'b1;
                            state <= STATE_IDLE;
                            finish <= 1'b1;
                        end
                    end
            endcase
        end
    end
endmodule